#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/mman.h>
#include <jpeglib.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/tcp.h>  // TCP_NODELAY
#include <time.h>
#include <errno.h>
#include <signal.h>
#include "v4l.h"
#include "opbcam-grab.h"

// A couple global-ish variables for remembering some stuff
static struct {
	void *start;
	size_t length;
} mmap_info;

int main(void)
{
	// Column of row pointers, which the JPEG compressor reads from
	JSAMPROW img_rows[SIZE_H];

	// File descriptors
	int udp, v4l, http;

	// JPEG stuff
	struct jpeg_error_mgr jerr;
	struct jpeg_compress_struct cinfo;

	int rv;
	int *img_start;
	time_t lastHeartbeat;

	printf("OPBCAM\n");

	mmap_info.start = NULL;

	initAlarm();

	// Listen for UDP packets
	if ((udp = initUDP()) < 0)
	{
		printf("Error: can't bind UDP socket.\n");
		return -1;
	}

	// Daemon loop
	while (1)
	{
		printf("Waiting for heartbeat...\n");
		if (recvUDP(udp, 1) < 0)
		{
			return -1;
		}
		lastHeartbeat = time(NULL);


		printf("Initializing Camera.\n");
		if ((v4l = initCam(img_rows)) < 0)
		{
			continue;
		}

		// Pretend that there's an integer overlapping the beginning
		// of the frame, so we can write a magic value, and check
		// to see if it's changed.
		img_start = (int *)img_rows[0];

		printf("Initializing HTTP stream.\n");
		if ((http = initHTTP()) < 0)
		{
			close(v4l);
			continue;
		}

		printf("Initializing JPEG compressor.\n");
		initJpeg(&cinfo, &jerr);
		printf("Setting up Jpeg Destination.\n");
		tcp_dest(&cinfo, http);

		printf("Capturing Video...\n");
		if (!V4LMCapture(v4l, 0)) while (1)
		{
			// Write the magic value
			*img_start = 0xDEADBEEF;

			// This alarm prevents lockups with some drivers
			// if you pull out the USB cable.
			alarm(3);
			rv = V4LMSync(v4l, 0);
			alarm(0);
			if (rv) break;

			// If the frame failed to update, then the
			// magic value will still be there 
			if (*img_start == 0xDEADBEEF) break;

			// Tell camera to start working on the next frame
			if (V4LMCapture(v4l, 0)) break;

			// JPEG-compress the frame we just got
			if (compressFrame(&cinfo, img_rows) < 0) break;

			// Check for incoming UDP packets to see if the
			// www server told us to start or stay running.
			// If nothing arrives for 30 seconds, then stop.
			rv = recvUDP(udp, 0);
			if (rv == 1)
			{
				lastHeartbeat = time(NULL);
			}
			else if (rv == 2 || time(NULL) - lastHeartbeat > 12)
			{
				break;
			}
		}
		jpeg_destroy_compress(&cinfo);
		close(http);

		/* Bad Things happen if we don't unmap this
		 * memory before closing the v4l device */
		if (mmap_info.start != NULL)
		{
			munmap(mmap_info.start, mmap_info.length);
			mmap_info.start = NULL;
		}
		close(v4l);
	}
}

int compressFrame(j_compress_ptr cinfo, JSAMPROW *img_rows)
{
	int i,j;
	unsigned char temp;

	// convert image from BGR -> RGB
	for (i=0; i<SIZE_H; i++)
	{
		for (j=0; j<SIZE_W*3; j+=3)
		{
			temp = img_rows[i][j];
			img_rows[i][j] = img_rows[i][j+2];
			img_rows[i][j+2] = temp;
		}
	}

	// Do jpeg compression, and transmit via HTTP
	jpeg_start_compress(cinfo, TRUE);
	jpeg_write_scanlines(cinfo, img_rows, SIZE_H);
	jpeg_finish_compress(cinfo);

	// Any problems?
	return tcp_dest_error ? -1 : 0;
}

int sendall(int s, char *buf, size_t len)
{
	// Send a big block of data
	int n;
	while (len > 0)
	{
		n = send(s, buf, len, 0);
		if (n == -1)
		{
			perror("send");
			return -1;
		}
		buf += n;
		len -= n;
		if (len > 0) usleep(1000);
	}
	return 0;
}

int initUDP()
{
	int fd;
	struct sockaddr_in my_addr;

	// Create UDP socket
	if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror("socket");
		return -1;
	}

	// Set up address/port
	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons(UDP_PORT);
	my_addr.sin_addr.s_addr = INADDR_ANY;
	memset(&(my_addr.sin_zero), '\0', 8);

	// Bind to local port
	if (bind(fd, (struct sockaddr *)&my_addr,
				sizeof(struct sockaddr)) < 0)
	{
		perror("bind");
		close(fd);
		return -1;
	}

	return fd;
}

int recvUDP(int fd, int block)
{
	// Return Value:
	//  -1 = Error
	//   0 = None
	//   1 = "ON"
	//   2 = "OFF"

	static socklen_t addr_len = sizeof(struct sockaddr_in);
	struct sockaddr_in their_addr;
	static char buf[UDP_BUF];
	int nbytes;
	int gotOn = 0;

	// enable or disable blocking
	fcntl(fd, F_SETFL, block ? 0 : O_NONBLOCK);

	/*
	 * In blocking mode, keep blocking until we receive the "ON"
	 * message.  In non-blocking mode, read until empty, then return
	 * whether or not the "ON" message was found
	 */

	while (1)
	{
		if ((nbytes = recvfrom(fd, buf, UDP_BUF, 0,
			(struct sockaddr *)&their_addr, &addr_len)) < 0)
		{
			if (!block && errno == EAGAIN)
			{
				// Not blocking, nothing waiting:
				// return status
				return gotOn;	
			}
			else
			{
				// Unexpected error
				perror("recvfrom");
				return -1;
			}
		}

		if (nbytes != 3) continue;

		/* TODO: Could verify the source address, but it's not that important */

		if (!strncmp("ON!", buf, 3))
		{
			printf("Received 'ON' Packet\n");
			gotOn = 1;
			if (block) return 1;
		}
		else if (!strncmp("OFF", buf, 3))
		{
			printf("Received 'OFF' Packet\n");
			if (!block) return 2;
		}
	}
}

int initCam(JSAMPROW *img_rows)
{
	int i, fd;
	V4LMMInfo mmInfo;
	V4LFormat format;
	char *mmap_buf;

	// Open the video device
	if ((fd = open(VIDEO_DEV, O_RDWR)) < 0)
	{
		perror(VIDEO_DEV);
		return -1;
	}

	// Figure out how it will be mmapped
	if (V4LMGetMMInfo(fd, &mmInfo)) 
	{
		perror("V4LGetMMInfo");
		close(fd);
		return -1;
	}

	/*
	printf("--MMAP Stuff--\n");
	printf("Size: %d\n", mmInfo.size);
	printf("Frames: %d\n\n", mmInfo.frames);
	*/

	// set up mmap to access the frames
	if ((mmap_buf = mmap(0, mmInfo.size, PROT_READ,
					MAP_SHARED, fd, 0)) == MAP_FAILED)
	{
		perror("mmap");
		close(fd);
		return -1;
	}

	// save the mmap info so we can unmap it before closing
	mmap_info.start = mmap_buf;
	mmap_info.length = mmInfo.size;

	// Set formatting
	format.width = SIZE_W;
	format.height = SIZE_H;
	format.format = V4L_FORMAT_RGB24;
	V4LMSetFormat(&format);

	// Build table of row pointers
	for (i=0; i<SIZE_H; i++)
	{
		img_rows[i] = (JSAMPROW)&mmap_buf[i * 3 * SIZE_W];
	}

	return fd;
}

int initHTTP(void)
{
	// Get the HTTP socket connected and ready
	struct hostent *he;
	struct sockaddr_in remote_addr;
	int flag = 1;
	int fd;


	// Build HTTP Request
	static char *REQUEST =
		"POST " HTTP_PATH " HTTP/1.0\r\n"
		"Host: " HTTP_HOST "\r\n"
		"Content-length: 2147483647\r\n\r\n";

	// Look up server's IP address
	if ((he=gethostbyname(HTTP_HOST)) == NULL)
	{
		perror("gethostbyname");
		return -1;
	}

	// Create socket for connecting to server
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror("socket");
		return -1;
	}

	// Set up target address
	remote_addr.sin_family = AF_INET;
	remote_addr.sin_port = htons(80);
	remote_addr.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&(remote_addr.sin_zero), 0, 8);

	// Reduce TCP buffering delay
	if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
				(char *)&flag, sizeof(int)) == -1)
	{
		perror("setsockopt");
		close(fd);
		return -1;
	}

	// Establish HTTP connection
	if (connect(fd, (struct sockaddr *)&remote_addr,
				sizeof(struct sockaddr)) == -1)
	{
		perror("connect");
		close(fd);
		return -1;
	}

	// Send the POST request
	if (sendall(fd, REQUEST, strlen(REQUEST)) == -1)
	{
		close(fd);
		return -1;
	}

	return fd;
}

void initJpeg(struct jpeg_compress_struct *cinfo, struct jpeg_error_mgr *jerr)
{
	// Set up the JPEG compressor
	cinfo->err = jpeg_std_error(jerr);
	jpeg_create_compress(cinfo);

	cinfo->image_width = SIZE_W;
	cinfo->image_height = SIZE_H;
	cinfo->input_components = 3;
	cinfo->in_color_space = JCS_RGB;
	jpeg_set_defaults(cinfo);

	// Set libjpeg to floating point mode (slightly faster)
	cinfo->dct_method = JDCT_FLOAT;
}

/* These 2 functions are for breaking out of the sync call when the
 * ioctl hangs.  Somewhat copied from xawtv */
void sigalarm(int signal)
{
	printf("Got an alarm. Sync failed?\n");
}
void initAlarm(void)
{
	struct sigaction act;
	memset(&act, 0, sizeof(act));
	act.sa_handler = sigalarm;
	sigemptyset(&act.sa_mask);
	sigaction(SIGALRM, &act, NULL);
}

/*
void getCaps(int v4l)
{
	V4LCaps caps;
	V4LMaskName *maskName;

	if (V4LGetCaps(v4l, &caps))
	{
		perror("V4lGetCaps");
	}

	printf("--Caps--\n");

	printf("Name: %s\n", caps.name);
	printf("Type: 0x%X\n", caps.type);

	for (maskName = V4LVideoTypeMaskList; maskName->mask; maskName++)
	{
		if (caps.type & maskName->mask)
			printf("\t%s\n", maskName->name);
	}

	printf("Channels: %d\n", caps.channels);
	printf("Audios: %d\n", caps.audios);
	printf("MaxWidth: %d\n", caps.maxwidth);
	printf("MaxHeight: %d\n", caps.maxheight);
	printf("MinWidth: %d\n", caps.minwidth);
	printf("MinHeight: %d\n\n", caps.minheight);
}

int getSource(int v4l)
{
	V4LChannel channel;
	if (V4LGetSource(v4l, 0, &channel))
	{
		perror("V4LGetSource");
	}

	printf("--Channel--\n");
	printf("Channel: %d\n", channel.channel);
	printf("Name: %s\n", channel.name);
	printf("Tuners: %d\n", channel.tuners);
	printf("Flags: 0x%X\n", channel.flags);
	printf("Type: %d (%s)\n", channel.type,
			V4LChannelTypeName(channel.type));
	printf("Norm: %d\n\n", channel.norm);
}
*/
