#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <string.h>
#include <time.h>
#include "opbcam-common.h"

#define BOUNDARY "NEXTFRAME"

static char *HTTP_RESPONSE =
	"HTTP/1.1 200 OK\r\n"
	"Content-type: multipart/x-mixed-replace; boundary=" BOUNDARY "\r\n"
	"Expires: Mon, 26 Jul 1997 05:00:00 GMT\r\n"
	"Cache-Control: no-store, no-cache, must-revalidate\r\n"
	"Cache-Control: post-check=0, pre-check=0\r\n"
	"Pragma: no-cache\r\n\r\n"
	"--" BOUNDARY "\r\n";

// Header/footer for normal frames
#define F_HEADER "Content-type: image/jpeg\r\n\r\n"
#define F_FOOTER "\r\n--" BOUNDARY "\r\n"

// Header/footer for the status images
#define S_HEADER "Content-type: image/png\r\n\r\n"
#define S_FOOTER "\r\n--" BOUNDARY

int get_vlock(int fd);
void ping_server(void);
void sendFrame(int fd, struct shm_data *shm);
void sendStatus(char *file, int final);

int main(void)
{
	int shm_fd;
	struct shm_data *shm = 0;
	int last_sync, new_sync;
	time_t chk_time;

	init_locks();

	// Connect to shared memory
	shm_fd = init_shm(&shm, PROT_READ);
	last_sync = shm->sync;

	// Grab a viewer lock, so we're countable.
	if (get_vlock(shm_fd) == -1)
	{
		return 0;
	}

	// Activate the server
	ping_server();

	// Send HTTP response header
	writeall(HTTP_RESPONSE, strlen(HTTP_RESPONSE));

	// Send "Loading..." image
	//sendStatus(LOADING_IMG, 0);

	while(1)
	{
		// Wait for frame update
		chk_time = time(NULL);
		while ((new_sync = shm->sync) == last_sync)
		{
			if (time(NULL)-chk_time >= TIMEOUT)
			{
				// Send "Offline" image
				sendStatus(OFFLINE_IMG, 1);
				return 0;
			}
			usleep(1000);
		}
		last_sync = new_sync;
		sendFrame(shm_fd, shm);
	}
}

int get_vlock(int fd)
{
	// Grab the first available viewer lock
	int i;
	struct flock fl;

	for (i=0; i<MAX_VIEWERS; i++)
	{
		make_lock(&fl, 0, 1);
		fl.l_start = i;
		fl.l_len = 1;
		fl.l_type = F_WRLCK;
		if (fcntl(fd, F_SETLK, &fl) == 0)
		{
			return 0;
		}
	}
	return -1;
}

void ping_server(void)
{
	int fd;
	struct sockaddr_in addr;

	fd = initUDP(&addr);
	sendUDP(fd, "ON!", 3, &addr);
	close(fd);
}

void sendFrame(int fd, struct shm_data *shm)
{
	char buf[MAX_FRAME+128];
	char *bufp = buf;
	int index, len;

	memcpy(bufp, F_HEADER, sizeof(F_HEADER)-1);
	bufp += sizeof(F_HEADER)-1;

	// lock the index, and read it
	lock_wait(fd, F_RDLCK, &index_lk);
	index = shm->index;

	// Check for bad index
	if (index < 0 || index > BUF_COUNT)
	{
		lock_wait(fd, F_UNLCK, &index_lk);
		return;
	}

	// lock the frame, unlock the index
	lock_wait(fd, F_RDLCK, &buf_lk[index]);
	lock_wait(fd, F_UNLCK, &index_lk);

	// Get frame length
	len = shm->buf[index].len;

	// Check for bad length
	if (len <=0 || len > MAX_FRAME)
	{
		lock_wait(fd, F_UNLCK, &buf_lk[index]);
		return;
	}

	// Get frame, unlock buffer
	memcpy(bufp, shm->buf[index].frame, len);
	lock_wait(fd, F_UNLCK, &buf_lk[index]);
	bufp += len;

	// Add footer
	memcpy(bufp, F_FOOTER, sizeof(F_FOOTER)-1);
	bufp += sizeof(F_FOOTER)-1;

	// Send it all
	writeall(buf, bufp-buf);
}

void sendStatus(char *file, int final)
{
	char buf[MAX_IMG+128];
	char *bufp = buf;
	int fd;

	if ((fd = open(file, O_RDONLY, 0)) == -1)
	{
		perror("open");
		exit(1);
	}

	memcpy(bufp, S_HEADER, sizeof(S_HEADER)-1);
	bufp += sizeof(S_HEADER)-1;
	bufp += read(fd, bufp, MAX_IMG);
	memcpy(bufp, S_FOOTER, sizeof(S_FOOTER)-1);
	bufp += sizeof(S_FOOTER)-1;
	if (final)
	{
		memcpy(bufp, "--\r\n", 4);
		bufp += 4;
	}
	else
	{
		memcpy(bufp, "\r\n", 2);
		bufp += 2;
	}
	writeall(buf, bufp-buf);
}
